import os
from typing import List

from sqlalchemy import func

from mlcomp import DATA_FOLDER, MODEL_FOLDER
from mlcomp.db.core import PaginatorOptions
from mlcomp.db.models import Project, Dag, Task
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.io import yaml_dump


class ProjectProvider(BaseDataProvider):
    model = Project

    def add_project(
            self,
            name: str,
            class_names: dict = None,
            sync_folders: List[str] = None,
            ignore_folders: List[str] = None
    ):
        class_names = class_names or {}
        ignore_folders = ignore_folders or []

        assert type(class_names) == dict, 'class_names type must be dict'
        assert isinstance(ignore_folders, list), \
            'ignore_folders type must be list'

        project = Project(
            name=name,
            class_names=yaml_dump(class_names),
            sync_folders=yaml_dump(sync_folders),
            ignore_folders=yaml_dump(ignore_folders),
        )
        project = self.session.add(project)

        os.makedirs(os.path.join(DATA_FOLDER, name), exist_ok=True)
        os.makedirs(os.path.join(MODEL_FOLDER, name), exist_ok=True)

        return project

    def edit_project(
            self,
            name: str,
            class_names: dict,
            sync_folders: List[str],
            ignore_folders: List[str]
    ):
        assert type(class_names) == dict, 'class_names type must be dict'
        assert isinstance(ignore_folders, list), \
            'ignore_folders type must be list'

        project = self.by_name(name)
        project.class_names = yaml_dump(class_names)
        project.sync_folders = yaml_dump(sync_folders)
        project.ignore_folders = yaml_dump(ignore_folders)
        self.commit()

    def get(self, filter: dict = None, options: PaginatorOptions = None):
        filter = filter or {}

        query = self.query(Project,
                           func.count(Dag.id),
                           func.max(Task.last_activity)). \
            join(Dag, Dag.project == Project.id, isouter=True). \
            join(Task, isouter=True). \
            group_by(Project.id)

        if filter.get('name'):
            query = query.filter(Project.name.like(f'%{filter["name"]}%'))

        total = query.count()
        paginator = self.paginator(query, options)
        res = []
        for p, dag_count, last_activity in paginator.all():
            last_activity = self.serializer.serialize_datetime(last_activity) \
                if last_activity else None

            file_size, img_size = self.query(func.sum(Dag.file_size),
                                             func.sum(Dag.img_size)).filter(
                Dag.project == p.id).one()

            res.append(
                {
                    'dag_count': dag_count,
                    'last_activity': last_activity,
                    'img_size': int(img_size or 0),
                    'file_size': int(file_size or 0),
                    'id': p.id,
                    'name': p.name,
                    'sync_folders': p.sync_folders,
                    'ignore_folders': p.ignore_folders,
                    'class_names': p.class_names
                })
        return {'total': total, 'data': res}

    def all_last_activity(self):
        query = self.query(Project,
                           func.max(Task.last_activity)). \
            join(Dag, Dag.project == Project.id, isouter=True). \
            join(Task, isouter=True). \
            group_by(Project.id)

        res = query.all()
        for p, last_activity in res:
            p.last_activity = last_activity
        return [r[0] for r in res]

    def by_name(self, name: str):
        return self.query(Project).filter(Project.name == name).first()


__all__ = ['ProjectProvider']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

# Based on https://github.com/catalyst-team/catalyst/blob/master/setup.py

import io
import os
import sys
from glob import glob
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'mlcomp'
DESCRIPTION = 'Machine learning pipelines. ' \
              'Especially, for competitions, like Kaggle'
URL = 'https://github.com/catalyst-team/mlcomp'
EMAIL = 'lightsanweb@gmail.com'
AUTHOR = 'Evgeny Semyonov'
REQUIRES_PYTHON = '>=3.6.0'

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))


def load_requirements():
    with open(os.path.join(PROJECT_ROOT, 'requirements.txt'), 'r') as f:
        return f.read()


def load_readme():
    readme_path = os.path.join(PROJECT_ROOT, 'README.md')
    with io.open(readme_path, encoding='utf-8') as f:
        return '\n' + f.read()


def load_version():
    context = {}
    with open(os.path.join(PROJECT_ROOT, 'mlcomp', '__version__.py')) as f:
        exec(f.read(), context)
    return context['__version__']


def files(directory):
    objs = glob(os.path.join(directory, '**'), recursive=True)
    folders = [o for o in objs if os.path.isdir(o)]
    for folder in folders:
        if '__pycache__' in folder:
            continue

        folder_files = [
            os.path.join(folder, f) for f in os.listdir(folder)
            if os.path.isfile(os.path.join(folder, f))
        ]
        yield folder, folder_files


def get_data_files():
    res = []

    front_folder = 'mlcomp/server/front/dist'
    front_readme = 'https://github.com/catalyst-team/' \
                   'mlcomp/blob/master/mlcomp/server/front/README.md'

    if not os.path.exists(front_folder):
        raise Exception(f'There is no {front_folder}. '
                        f'You should compile front-end files first'
                        f' Please check {front_readme}')
    if len(os.listdir(front_folder)) == 0:
        raise Exception(f'There are no files in {front_folder}. '
                        f'You should compile front-end files first'
                        f' Please check {front_readme}')

    res.extend(files('mlcomp/utils'))
    res.extend(files('mlcomp/bin'))
    res.extend(files('mlcomp/docker'))
    res.extend(files(front_folder))
    res.extend(files('mlcomp/migration'))
    return res


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(PROJECT_ROOT, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(
            '{0} setup.py sdist bdist_wheel --universal'.format(
                sys.executable
            )
        )

        self.status('Uploading the package to PyPI via Twine…')
        os.system('python -m twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(load_version()))
        os.system('git push --tags')

        sys.exit()


setup(
    name=NAME,
    version=load_version(),
    description=DESCRIPTION,
    long_description=load_readme(),
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=load_requirements(),
    include_package_data=True,
    data_files=get_data_files(),
    zip_safe=False,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand
    },
    entry_points={
        'console_scripts': [
            'mlcomp-server = mlcomp.server.__main__:main',
            'mlcomp-worker = mlcomp.worker.__main__:main',
            'mlcomp-contrib = mlcomp.contrib.__main__:main',
            'mlcomp = mlcomp.__main__:main',
        ],
    }
)



import json
import shutil
from enum import Enum
import os
import time

import socket

from kaggle.models import DatasetNewRequest

from mlcomp.db.core import Session
from mlcomp.db.enums import ComponentType
from mlcomp.db.providers import ModelProvider
from mlcomp.worker.executors.base.equation import Equation
from mlcomp.worker.executors.base.executor import Executor
from mlcomp.utils.logging import create_logger
from mlcomp.utils.config import Config

try:
    from kaggle import api
except OSError:
    logger = create_logger(Session.create_session(), __name__)
    logger.warning(
        'Could not find kaggle.json. '
        'Kaggle executors can not be used', ComponentType.Worker,
        socket.gethostname()
    )


class DownloadType(Enum):
    Kaggle = 0
    Link = 1


@Executor.register
class Download(Executor):
    def __init__(
            self,
            output: str,
            type=DownloadType.Kaggle,
            competition: str = None,
            link: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        if type == DownloadType.Kaggle and competition is None:
            raise Exception('Competition is required for Kaggle')
        self.type = type
        self.competition = competition
        self.link = link
        self.output = output

    def work(self):
        api.competition_download_files(self.competition, self.output)

    @classmethod
    def _from_config(
            cls, executor: dict, config: Config, additional_info: dict
    ):
        output = os.path.join(config.data_folder, config.get('output', '.'))
        return cls(output=output, competition=executor['competition'])


@Executor.register
class Submit(Equation):
    def __init__(
            self,
            competition: str,
            submit_type: str = 'file',
            predict_column: str = None,
            kernel_suffix: str = 'api',
            message: str = '',
            wait_seconds: int = 60 * 20,
            file: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.competition = competition
        self.wait_seconds = wait_seconds
        self.submit_type = submit_type
        self.kernel_suffix = kernel_suffix
        self.predict_column = predict_column
        self.message = message or f'model_id = {self.model_id}'

        default_file = f'data/submissions/{self.model_name}_{self.suffix}.csv'
        self.file = file or default_file
        self.file_name = os.path.basename(self.file)

        assert self.submit_type in ['file', 'kernel']
        if self.submit_type == 'kernel':
            assert self.predict_column, 'predict_column must be specified'

    def file_submit(self):
        self.info(f'file_submit. file = {self.file} start')
        api.competition_submit(
            self.file, message=self.message, competition=self.competition
        )
        self.info(f'file_submit. file = {self.file} end')

    def kernel_submit(self):
        self.info('kernel_submit updating dataset')

        folder = 'submit'
        os.makedirs(folder, exist_ok=True)

        shutil.copy(self.file, os.path.join(folder, self.file_name))

        config = api.read_config_file()
        username = config['username']
        title = f'{self.competition}-{self.kernel_suffix}-dataset'
        dataset_meta = {
            'title': f'{self.competition}-{self.kernel_suffix}-dataset',
            'id': f'{username}/{title}',
            'licenses': [{
                'name': 'CC0-1.0'
            }]
        }
        with open(f'{folder}/dataset-metadata.json', 'w') as f:
            json.dump(dataset_meta, f)

        res = api.dataset_status(dataset_meta['id'])
        if res != 'ready':
            res = api.dataset_create_new(folder=folder)
            if res.status == 'error':
                raise Exception('dataset_create_new Error: ' + res.error)

        res = api.dataset_create_version(folder, 'Updated')
        if res.status == 'error':
            raise Exception('dataset_create_version Error: ' + res.error)

        self.info('dataset updated')

        seconds_to_sleep = 20
        self.info(f'sleeping {seconds_to_sleep} seconds')
        time.sleep(seconds_to_sleep)

        slug = f'{self.competition}-{self.kernel_suffix}'
        kernel_meta = {
            'id': f'{username}/{slug}',
            'title': f'{self.competition}-{self.kernel_suffix}',
            'code_file': 'code.py',
            'language': 'python',
            'kernel_type': 'script',
            'is_private': 'true',
            'enable_gpu': 'false',
            'enable_internet': 'false',
            'dataset_sources': [dataset_meta['id']],
            'competition_sources': [self.competition],
            'name': f'{self.competition}-{self.kernel_suffix}'
        }
        with open(f'{folder}/kernel-metadata.json', 'w') as f:
            json.dump(kernel_meta, f)

        code = """
import pandas as pd

DATA_DIR = '../input/{self.competition}'
CSV_FILE = '../input/{self.competition}-' + \
           '{self.kernel_suffix}-dataset/{self.file_name}'

df = pd.read_csv(DATA_DIR + '/sample_submission.csv')
df_predict = pd.read_csv(CSV_FILE)

keys = [c for c in df.columns if c!='{self.predict_column}']
predict_values = dict()
for index, row in df_predict.iterrows():
    key = tuple([row[k] for k in keys])
    predict_values[key] = row

res = []
for index, row in df.iterrows():
    key = tuple([row[k] for k in keys])
    if key in predict_values:
        res.append(predict_values[key])
    else:
        res.append(row)

res = pd.DataFrame(res)
res.to_csv('submission.csv', index=False)
        """.replace('{self.competition}', self.competition).replace(
            '{self.kernel_suffix}', self.kernel_suffix
        ).replace('{self.file_name}', self.file_name
                  ).replace('{self.predict_column}', self.predict_column)

        with open(f'{folder}/code.py', 'w') as f:
            f.write(code)

        self.info('kernel data created')
        api.kernels_push(folder)
        self.info('kernel is pushed. waiting for the end of the commit')
        self.info(f'kernel address: https://www.kaggle.com/{username}/{slug}')

        seconds = self.wait_seconds
        for i in range(seconds):
            response = api.kernel_status(username, slug)
            if response['status'] == 'complete':
                self.info(f'kernel has completed successfully. '
                          f'Please go to '
                          f'https://www.kaggle.com/{username}/{slug} '
                          f'and push the button "Submit to the competition"')
                return
            if response['status'] == 'error':
                raise Exception(
                    f'Kernel is failed. Msg = {response["failureMessage"]}'
                )
            time.sleep(1)
            self.wait_seconds -= 1

        self.info(f'kernel is not finished after {seconds}')

    def work(self):
        submissions = api.competition_submissions(self.competition)
        submission_refs = {s.ref for s in submissions}

        if self.submit_type == 'file':
            self.file_submit()
        else:
            self.kernel_submit()

        self.info('waiting for the submission on Kaggle')

        step = 10
        for i in range(int(self.wait_seconds // step)):
            try:
                submissions = api.competition_submissions(self.competition)
                for s in submissions:
                    if s.ref not in submission_refs:
                        if s.status == 'complete':
                            if s.publicScore is None:
                                raise Exception(
                                    'Submission is complete, '
                                    'but publicScore is None'
                                )
                            score = float(s.publicScore)
                            if self.model_id:
                                provider = ModelProvider(self.session)
                                model = provider.by_id(self.model_id)
                                model.score_public = score
                                provider.commit()

                            return {'res': score}
                        elif s.status == 'error':
                            raise Exception(
                                f'Submission error '
                                f'on Kaggle: {s.errorDescription}'
                            )

                        break
            except TypeError:
                pass

            time.sleep(step)
        raise Exception(
            f'Submission is not '
            f'complete after {self.wait_seconds}'
        )


__all__ = ['Download', 'Submit']

if __name__ == '__main__':
    submit = Submit(
        competition='severstal-steel-defect-detection',
        name='resnetunet',
        submit_type='kernel',
        predict_column='EncodedPixels'
    )
    submit.work()



from .decoder import FPNDecoder
from ..base import EncoderDecoder
from ..encoders import get_encoder


class FPN(EncoderDecoder):
    """FPN_ is a fully convolution neural
    network for image semantic segmentation
    Args:
        encoder_name: name of classification model
         (without last dense layers) used as feature
                extractor to build segmentation model.
        encoder_weights: one of ``None`` (random initialization),
        ``imagenet`` (pre-training on ImageNet).
        decoder_pyramid_channels: a number of convolution
        filters in Feature Pyramid of FPN_.
        decoder_segmentation_channels: a number of
        convolution filters in segmentation head of FPN_.
        classes: a number of classes for output
        (output shape - ``(batch, classes, h, w)``).
        dropout: spatial dropout rate in range (0, 1).
        activation: activation function used
        in ``.predict(x)`` method for inference.
            One of [``sigmoid``, ``softmax``, callable, None]

    Returns:
        ``torch.nn.Module``: **FPN**

    .. _FPN:
        http://presentations.cocodataset.org/COCO17-Stuff-FAIR.pdf

    """

    def __init__(
            self,
            encoder_name='resnet34',
            encoder_weights='imagenet',
            decoder_pyramid_channels=256,
            decoder_segmentation_channels=128,
            classes=1,
            dropout=0.2,
            activation='sigmoid',
    ):
        encoder = get_encoder(
            encoder_name,
            encoder_weights=encoder_weights
        )

        decoder = FPNDecoder(
            encoder_channels=encoder.out_shapes,
            pyramid_channels=decoder_pyramid_channels,
            segmentation_channels=decoder_segmentation_channels,
            final_channels=classes,
            dropout=dropout,
        )

        super().__init__(encoder, decoder, activation)

        self.name = 'fpn-{}'.format(encoder_name)



from collections import OrderedDict

from mlcomp.db.report_info.item import ReportLayoutItem


class ReportLayoutImgSegment(ReportLayoutItem):
    def __init__(
        self,
        name: str,
        max_height: int,
        max_width: int
    ):
        super().__init__(name)

        self.max_height = max_height
        self.max_width = max_width

    @classmethod
    def from_dict(cls, name: str, value: OrderedDict):
        value.pop('type')
        max_height = value.pop('max_height', None)
        max_width = value.pop('max_width', None)

        assert len(value) == 0, f'Unknown parameter in ' \
            f'report.img_segment={value.popitem()}'

        return cls(
            name,
            max_height=max_height,
            max_width=max_width
        )


__all__ = ['ReportLayoutImgSegment']



from typing import List
from itertools import groupby

from mlcomp.db.models import ReportSeries, Task
from mlcomp.db.providers.base import BaseDataProvider


class ReportSeriesProvider(BaseDataProvider):
    model = ReportSeries

    def by_dag(self, dag: int, metrics: List[str]):
        tasks = self.query(Task.id, Task.name).filter(Task.dag == dag).all()
        ids = [id for id, _ in tasks]

        series = self.query(ReportSeries).filter(
            ReportSeries.task.in_(ids)).filter(
            ReportSeries.name.in_(metrics)).order_by(ReportSeries.task).all()

        res = []
        for task_id, task_series in groupby(series, key=lambda x: x.task):
            task_series = list(task_series)
            task_series = sorted(task_series, key=lambda x: x.name)

            for name, name_series in groupby(task_series,
                                             key=lambda x: x.name):
                name_series = list(name_series)
                name_series = sorted(name_series, key=lambda x: x.part)
                groups = []
                for group, part_series in groupby(name_series,
                                                  key=lambda x: x.part):
                    part_series = list(part_series)
                    part_series = sorted(part_series, key=lambda x: x.epoch)
                    group = {'name': group,
                             'epoch': [s.epoch for s in part_series],
                             'value': [s.value for s in part_series]
                             }
                    groups.append(group)
                res.append((task_id, name, groups))
        return res


__all__ = ['ReportSeriesProvider']



from collections import OrderedDict


class ReportLayoutItem:
    def __init__(self, name: str):
        self.name = name

    @classmethod
    def from_dict(cls, name: str, value: OrderedDict):
        value.pop('type')
        assert len(value) == 0, f'Unknown parameter in ' \
            f'report info item = {name}: {value.popitem()}'
        return cls(name)


__all__ = ['ReportLayoutItem']



from os.path import join

import cv2
import numpy as np

from mlcomp.contrib.dataset.classify import ImageDataset


class ImageWithMaskDataset(ImageDataset):
    def __init__(
            self, *, mask_folder: str, crop_positive=None, encode=True,
            include_binary=False, **kwargs
    ):
        assert mask_folder, 'Mask folder is required'
        self.mask_folder = mask_folder
        self.crop_positive = crop_positive
        self.encode = encode
        self.include_binary = include_binary
        if not encode and kwargs.get('num_classes', 1) > 1:
            kwargs['num_classes'] += 1

        super().__init__(**kwargs)

    def preprocess_row(self, row: dict):
        row['image'] = join(self.img_folder, row['image'])
        row['mask'] = join(self.mask_folder, row['mask'])

    def _get_item_before_transform(self, row: dict, item: dict):
        if 'mask' in row and self.mask_folder:
            item['mask'] = self.read_image_file(row['mask'], True)
            self._process_crop_positive(item)

    def _get_item_after_transform(
            self, row: dict, transformed: dict, res: dict
    ):
        if 'mask' in transformed:
            mask = transformed['mask'].astype(np.int64)
            if len(mask.shape) == 2:
                mask_encoded = np.zeros(
                    (self.num_classes, *mask.shape), dtype=np.float32
                )
                if self.num_classes == 1:
                    mask = (mask >= 1).astype(np.uint8)

                if self.encode:
                    for i in range(1, self.num_classes + 1):
                        mask_encoded[i - 1] = mask == i

                    mask = mask_encoded
                else:
                    for i in range(1, self.num_classes + 1):
                        mask_encoded[i - 1] = mask == i

                    res['targets_encoded'] = mask_encoded

            res['targets'] = mask.astype(np.float32)

            if self.include_binary:
                for i, c in enumerate(mask):
                    res[f'empty_{i}'] = int(c.sum() == 0)
                res['empty_all'] = int(mask.sum() == 0)

    def _process_crop_positive(self, item: dict):
        if not self.crop_positive:
            return
        mask = item['mask']
        size = mask.shape

        # import matplotlib.pyplot as plt
        # plt.imshow(item['image'])
        # plt.show()
        # plt.imshow(item['mask'] * 53)
        # plt.show()
        crop_pos_y = self.crop_positive[0]
        if type(crop_pos_y) == tuple:
            crop_pos_y = np.random.randint(crop_pos_y[0], crop_pos_y[1])

        crop_pos_x = self.crop_positive[1]
        if type(crop_pos_x) == tuple:
            crop_pos_x = np.random.randint(crop_pos_x[0], crop_pos_x[1])

        p = self.crop_positive[2]

        if mask.sum() == 0 or np.random.uniform(0, 1) < p:
            min_y = 0
            max_y = size[0] - crop_pos_y
            min_x = 0
            max_x = size[1] - crop_pos_x
        else:
            mask_binary = (mask > 0).astype(np.uint8)
            contours, _ = cv2.findContours(
                mask_binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
            )
            contour = contours[np.random.randint(0, len(contours))]
            rect = cv2.boundingRect(contour)
            min_y = rect[1] - crop_pos_y + int(rect[3] // 2)
            max_y = rect[1] + int(rect[3] // 2)

            min_x = rect[0] - crop_pos_x + int(rect[2] // 2)
            max_x = rect[0] + int(rect[2] // 2)

        min_x = max(0, min_x)
        min_y = max(0, min_y)

        max_x = min(max_x, size[1] - crop_pos_x)
        max_y = min(max_y, size[0] - crop_pos_y)

        x = np.random.randint(min_x, max_x + 1)
        y = np.random.randint(min_y, max_y + 1)
        item['image'] = item['image'][y:y + crop_pos_y, x:x + crop_pos_x]

        item['mask'] = item['mask'][y:y + crop_pos_y, x:x + crop_pos_x]

        # plt.imshow(item['image'])
        # plt.show()
        # plt.imshow(item['mask'] * 53)
        # plt.show()



from .img import ReportImgProvider
from .report import ReportProvider
from .layout import ReportLayoutProvider
from .series import ReportSeriesProvider
from .task import ReportTasksProvider

__all__ = [
    'ReportImgProvider', 'ReportProvider', 'ReportLayoutProvider',
    'ReportSeriesProvider', 'ReportTasksProvider'
]



import time
import socket
import json
import os
import traceback
from multiprocessing import cpu_count

import click
import GPUtil
import psutil
import numpy as np
import torch

from mlcomp.report import check_statuses
from mlcomp.utils.io import yaml_load

from mlcomp import ROOT_FOLDER, MASTER_PORT_RANGE, CONFIG_FOLDER, \
    DOCKER_IMG, DOCKER_MAIN, IP, PORT, WORKER_USAGE_INTERVAL, \
    SYNC_WITH_THIS_COMPUTER, CAN_PROCESS_TASKS
from mlcomp.db.core import Session
from mlcomp.db.enums import ComponentType, TaskStatus
from mlcomp.utils.logging import create_logger
from mlcomp.db.providers import DockerProvider, TaskProvider
from mlcomp.utils.schedule import start_schedule
from mlcomp.utils.misc import dict_func, now, disk, get_username, \
    kill_child_processes, get_pid
from mlcomp.worker.app import app
from mlcomp.db.providers import ComputerProvider
from mlcomp.db.models import ComputerUsage, Computer, Docker
from mlcomp.utils.misc import memory
from mlcomp.worker.sync import FileSync

_session = Session.create_session(key='worker')


@click.group()
def main():
    pass


def error_handler(f):
    name = f.__name__
    wrapper_vars = {'session': Session.create_session(key=name)}
    wrapper_vars['logger'] = create_logger(wrapper_vars['session'], name)

    hostname = socket.gethostname()

    def wrapper():
        try:
            f(wrapper_vars['session'], wrapper_vars['logger'])
        except Exception as e:
            if Session.sqlalchemy_error(e):
                Session.cleanup(name)

                wrapper_vars['session'] = Session.create_session(key=name)
                wrapper_vars['logger'] = create_logger(wrapper_vars['session'],
                                                       name)

            wrapper_vars['logger'].error(
                traceback.format_exc(), ComponentType.WorkerSupervisor,
                hostname
            )

    return wrapper


@error_handler
def stop_processes_not_exist(session: Session, logger):
    provider = TaskProvider(session)
    hostname = socket.gethostname()
    tasks = provider.by_status(
        TaskStatus.InProgress,
        task_docker_assigned=DOCKER_IMG,
        computer_assigned=hostname
    )
    # Kill processes which does not exist
    hostname = socket.gethostname()
    for t in tasks:
        if not psutil.pid_exists(t.pid):
            # tasks can retry the execution
            if (now() - t.last_activity).total_seconds() < 30:
                continue

            os.system(f'kill -9 {t.pid}')
            t.status = TaskStatus.Failed.value
            logger.error(
                f'process with pid = {t.pid} does not exist. '
                f'Set task to failed state',
                ComponentType.WorkerSupervisor, hostname, t.id
            )

            provider.commit()

            additional_info = yaml_load(t.additional_info)
            for p in additional_info.get('child_processes', []):
                logger.info(f'killing child process = {p}')
                os.system(f'kill -9 {p}')

    # Kill processes which exist but should not
    processes = get_pid('worker ')
    ids = [p['PID'] for p in processes]
    tasks = provider.by_ids(ids)
    tasks = {t.pid: t for t in tasks}

    for p in processes:
        pid = p['PID']
        if pid in tasks:
            task = tasks[pid]
            if task.status in [TaskStatus.Stopped.value,
                               TaskStatus.Failed.value,
                               TaskStatus.Skipped.value]:

                logger.info(f'Kill processes which exist but should not. '
                            f'Pid = {pid}')
                os.system(f'kill -9 {pid}')


@error_handler
def worker_usage(session: Session, logger):
    provider = ComputerProvider(session)
    docker_provider = DockerProvider(session)

    computer = socket.gethostname()
    docker = docker_provider.get(computer, DOCKER_IMG)
    usages = []

    count = int(10 / WORKER_USAGE_INTERVAL)
    count = max(1, count)

    for _ in range(count):
        # noinspection PyProtectedMember
        memory = dict(psutil.virtual_memory()._asdict())

        try:
            gpus = GPUtil.getGPUs()
        except ValueError as err:
            logger.info(f"Active GPUs not found: {err}")
            gpus = []

        usage = {
            'cpu': psutil.cpu_percent(),
            'disk': disk(ROOT_FOLDER)[1],
            'memory': memory['percent'],
            'gpu': [
                {
                    'memory': 0 if np.isnan(
                        g.memoryUtil) else g.memoryUtil * 100,
                    'load': 0 if np.isnan(g.load) else g.load * 100
                } for g in gpus
            ]
        }

        provider.current_usage(computer, usage)
        usages.append(usage)
        docker.last_activity = now()
        docker_provider.update()

        time.sleep(WORKER_USAGE_INTERVAL)

    usage = json.dumps({'mean': dict_func(usages, np.mean)})
    provider.add(ComputerUsage(computer=computer, usage=usage, time=now()))


@main.command()
@click.argument('number', type=int)
def worker(number):
    """
    Start worker

    :param number: worker index
    """
    name = f'{socket.gethostname()}_{DOCKER_IMG}'
    argv = [
        'worker', '--loglevel=INFO', '-P=solo', f'-n={name}_{number}',
        '-O fair', '-c=1', '--prefetch-multiplier=1', '-Q', f'{name},'
                                                            f'{name}_{number}'
    ]
    app.worker_main(argv)


@main.command()
@click.option('--workers', type=int, default=cpu_count(),
              help='count of workers')
def worker_supervisor(workers: int):
    """
    Start worker supervisor.
    This program controls workers ran on the same machine.
    Also, it writes metric of resources consumption.
    """
    check_statuses()

    host = socket.gethostname()

    logger = create_logger(_session, 'worker_supervisor')
    logger.info('worker_supervisor start',
                ComponentType.WorkerSupervisor,
                host)

    _create_computer(workers)
    _create_docker()

    start_schedule([(stop_processes_not_exist, 10)])

    if DOCKER_MAIN:
        syncer = FileSync()
        start_schedule([(worker_usage, 0)])
        start_schedule([(syncer.sync, 0)])

    name = f'{host}_{DOCKER_IMG}_supervisor'
    argv = [
        'worker', '--loglevel=INFO', '-P=solo', f'-n={name}', '-O fair',
        '-c=1', '--prefetch-multiplier=1', '-Q', f'{name}'
    ]

    logger.info('worker_supervisor run celery',
                ComponentType.WorkerSupervisor,
                host)

    app.worker_main(argv)


@main.command()
@click.option('--daemon', type=bool, default=False,
              help='start supervisord in a daemon mode')
@click.option('--debug', type=bool, default=False,
              help='use source files instead the installed library')
@click.option('--workers', type=int, default=cpu_count(),
              help='count of workers')
@click.option('--log_level', type=str, default='DEBUG',
              help='log level of supervisord')
def start(daemon: bool, debug: bool, workers: int, log_level: str):
    """
       Start worker_supervisor and workers
    """
    check_statuses()

    # creating supervisord config
    supervisor_command = 'mlcomp-worker worker-supervisor'
    worker_command = 'mlcomp-worker worker'
    if debug:
        supervisor_command = 'python mlcomp/worker/__main__.py ' \
                             'worker-supervisor'
        worker_command = 'python mlcomp/worker/__main__.py worker'

    daemon_text = 'false' if daemon else 'true'
    text = [
        '[supervisord]',
        f'nodaemon={daemon_text}',
        '',
        '[program:supervisor]',
        f'command={supervisor_command} --workers {workers}',
        'autostart=true',
        'autorestart=true',
        ''
    ]
    for p in range(workers):
        text.append(f'[program:worker{p}]')
        text.append(f'command={worker_command} {p}')
        text.append('autostart=true')
        text.append('autorestart=true')
        text.append('')

    conf = os.path.join(CONFIG_FOLDER, 'supervisord.conf')
    with open(conf, 'w') as f:
        f.writelines('\n'.join(text))

    os.system(f'supervisord ' f'-c {conf} -e {log_level}')


@main.command()
def stop():
    """
    Stop supervisord started by start command
    """
    check_statuses()

    lines = os.popen('ps -ef | grep supervisord').readlines()
    for line in lines:
        if 'mlcomp/configs/supervisord.conf' not in line:
            continue
        pid = int(line.split()[1])
        kill_child_processes(pid)


def _create_docker():
    docker = Docker(
        name=DOCKER_IMG,
        computer=socket.gethostname(),
        ports='-'.join(list(map(str, MASTER_PORT_RANGE))),
        last_activity=now()
    )
    DockerProvider(_session).create_or_update(docker, 'name', 'computer')


def _create_computer(workers: int):
    tot_m, used_m, free_m = memory()
    tot_d, used_d, free_d = disk(ROOT_FOLDER)
    computer = Computer(
        name=socket.gethostname(),
        gpu=torch.cuda.device_count(),
        cpu=workers,
        memory=tot_m,
        ip=IP,
        port=PORT,
        user=get_username(),
        disk=tot_d,
        root_folder=ROOT_FOLDER,
        sync_with_this_computer=SYNC_WITH_THIS_COMPUTER,
        can_process_tasks=CAN_PROCESS_TASKS
    )
    ComputerProvider(_session).create_or_update(computer, 'name')


if __name__ == '__main__':
    main()



from torch import nn

import mlcomp.contrib.segmentation as smb


class SegmentationModelPytorch(nn.Module):
    def __init__(
        self,
        arch: str,
        encoder: str,
        num_classes: int = 1,
        encoder_weights: str = 'imagenet',
        activation=None,
        **kwargs
    ):
        super().__init__()

        model = getattr(smb, arch)
        self.model = model(
            encoder_name=encoder,
            classes=num_classes,
            encoder_weights=encoder_weights,
            activation=activation,
            **kwargs
        )

    def forward(self, x):
        res = self.model.forward(x)
        if self.model.activation:
            res = self.model.activation(res)
        return res


__all__ = ['SegmentationModelPytorch']



import sqlalchemy as sa
from sqlalchemy import ForeignKey

from mlcomp.db.models.base import Base


class Log(Base):
    __tablename__ = 'log'

    id = sa.Column(sa.Integer, primary_key=True)
    step = sa.Column(sa.Integer, ForeignKey('step.id'))
    message = sa.Column(sa.String)
    time = sa.Column(sa.DateTime)
    level = sa.Column(sa.Integer)
    component = sa.Column(sa.Integer)
    module = sa.Column(sa.String)
    line = sa.Column(sa.Integer)
    task = sa.Column(sa.Integer, ForeignKey('task.id'))
    computer = sa.Column(sa.String, ForeignKey('computer.name'))


__all__ = ['Log']



import sqlalchemy as sa
import sqlalchemy.orm.session as session
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy import event

from mlcomp import SA_CONNECTION_STRING, DB_TYPE
from mlcomp.utils.misc import adapt_db_types


class Session(session.Session):
    __session = dict()

    def __init__(self, *args, **kwargs):
        key = kwargs.pop('key')
        if key in self.__session:
            raise Exception('Use static create_session for session creating')
        super().__init__(*args, **kwargs)

    @staticmethod
    def create_session(*, connection_string: str = None, key='default'):
        if key in Session.__session:
            return Session.__session[key][0]

        session_factory = scoped_session(sessionmaker(class_=Session, key=key))
        connect_args = {}
        if DB_TYPE == 'SQLITE':
            connect_args = {
                'check_same_thread': False,
                'timeout': 30
            }

        engine = sa.create_engine(
            connection_string or SA_CONNECTION_STRING,
            echo=False,
            connect_args=connect_args
        )
        if DB_TYPE == 'SQLITE':
            def _fk_pragma_on_connect(dbapi_con, con_record):
                dbapi_con.execute('pragma foreign_keys=ON')

            event.listen(engine, 'connect', _fk_pragma_on_connect)

        session_factory.configure(bind=engine)
        s = session_factory()

        Session.__session[key] = [s, engine]
        return s

    @classmethod
    def cleanup(cls, key: str):
        if key in cls.__session:
            s, engine = cls.__session[key]
            try:
                s.close()
            except Exception:
                pass

            try:
                engine.dispose()
            except Exception:
                pass

            del cls.__session[key]

    def query(self, *entities, **kwargs):
        try:
            return super().query(*entities, **kwargs)
        except Exception as e:
            self.rollback()
            raise e

    def add_all(self, objs, commit=True):
        try:
            for obj in objs:
                adapt_db_types(obj)
            super().add_all(objs)
        except Exception as e:
            raise e

        if commit:
            try:
                self.commit()
            except Exception as e:
                self.rollback()
                raise e

    def add(self, obj, commit=True, _warn=False):
        try:
            adapt_db_types(obj)
            super().add(obj, _warn=_warn)
        except Exception as e:
            raise e

        if commit:
            try:
                self.commit()
            except Exception as e:
                self.rollback()
                raise e
        return obj

    def commit(self):
        try:
            super().commit()
        except Exception as e:
            self.rollback()
            raise e

    def update(self):
        try:
            self.commit()
        except Exception as e:
            self.rollback()
            raise e

    @staticmethod
    def sqlalchemy_error(e):
        s = str(type(e))
        return 'sqlalchemy.' in s


__all__ = ['Session']



import re
import torch.nn as nn

from pretrainedmodels.models.torchvision_models import pretrained_settings
from torchvision.models.densenet import DenseNet


class DenseNetEncoder(DenseNet):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pretrained = False
        del self.classifier
        self.initialize()

    @staticmethod
    def _transition(x, transition_block):
        for module in transition_block:
            x = module(x)
            if isinstance(module, nn.ReLU):
                skip = x
        return x, skip

    def initialize(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out',
                                        nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):

        x = self.features.conv0(x)
        x = self.features.norm0(x)
        x = self.features.relu0(x)
        x0 = x

        x = self.features.pool0(x)
        x = self.features.denseblock1(x)
        x, x1 = self._transition(x, self.features.transition1)

        x = self.features.denseblock2(x)
        x, x2 = self._transition(x, self.features.transition2)

        x = self.features.denseblock3(x)
        x, x3 = self._transition(x, self.features.transition3)

        x = self.features.denseblock4(x)
        x4 = self.features.norm5(x)

        features = [x4, x3, x2, x1, x0]
        return features

    def load_state_dict(self, state_dict):
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))'
            r'\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]

        # remove linear
        state_dict.pop('classifier.bias')
        state_dict.pop('classifier.weight')

        super().load_state_dict(state_dict)


densenet_encoders = {
    'densenet121': {
        'encoder': DenseNetEncoder,
        'pretrained_settings': pretrained_settings['densenet121'],
        'out_shapes': (1024, 1024, 512, 256, 64),
        'params': {
            'num_init_features': 64,
            'growth_rate': 32,
            'block_config': (6, 12, 24, 16),
        }
    },

    'densenet169': {
        'encoder': DenseNetEncoder,
        'pretrained_settings': pretrained_settings['densenet169'],
        'out_shapes': (1664, 1280, 512, 256, 64),
        'params': {
            'num_init_features': 64,
            'growth_rate': 32,
            'block_config': (6, 12, 32, 32),
        }
    },

    'densenet201': {
        'encoder': DenseNetEncoder,
        'pretrained_settings': pretrained_settings['densenet201'],
        'out_shapes': (1920, 1792, 512, 256, 64),
        'params': {
            'num_init_features': 64,
            'growth_rate': 32,
            'block_config': (6, 12, 48, 32),
        }
    },

    'densenet161': {
        'encoder': DenseNetEncoder,
        'pretrained_settings': pretrained_settings['densenet161'],
        'out_shapes': (2208, 2112, 768, 384, 96),
        'params': {
            'num_init_features': 96,
            'growth_rate': 48,
            'block_config': (6, 12, 36, 24),
        }
    },

}



import numpy as np


def dice_numpy(targets, outputs, threshold=None, min_area=None,
               empty_one: bool = True, eps=1e-6):
    if threshold is not None:
        # noinspection PyUnresolvedReferences
        outputs = (outputs >= threshold).astype(np.uint8)

    targets_sum = targets.sum()
    outputs_sum = outputs.sum()

    if min_area and outputs_sum < min_area:
        outputs = np.zeros(outputs.shape, dtype=np.uint8)
        outputs_sum = 0

    if empty_one and targets_sum == 0 and outputs_sum == 0:
        return 1

    intersection = (targets * outputs).sum()
    union = targets_sum + outputs_sum
    dice = 2 * intersection / (union + eps)
    return dice



from typing import List
import os
import ast
from glob import glob
import pathspec
import pkg_resources

from mlcomp.db.core import Session
from mlcomp.utils.logging import create_logger
from mlcomp.utils.io import read_lines

_mapping = {
    'cv2': 'opencv-python',
    'sklearn': 'scikit-learn',
    'migrate': 'sqlalchemy-migrate'
}


def find_imports(
    path: str,
    files: List[str] = None,
    exclude_patterns: List[str] = None,
    encoding='utf-8'
):
    res = []
    raw_imports = []
    files = files if files is not None \
        else glob(os.path.join(path, '**', '*.py'), recursive=True)

    exclude_patterns = exclude_patterns \
        if exclude_patterns is not None else []
    spec = pathspec.PathSpec.from_lines(
        pathspec.patterns.GitWildMatchPattern, exclude_patterns
    )

    for file in files:
        if not file.endswith('.py'):
            continue
        file_rel = os.path.relpath(file, path)
        if spec.match_file(file_rel):
            continue

        with open(file, 'r', encoding=encoding) as f:
            content = f.read()
            try:
                tree = ast.parse(content)
                for node in ast.walk(tree):
                    if isinstance(node, ast.Import):
                        for subnode in node.names:
                            raw_imports.append((subnode.name, file_rel))
                    elif isinstance(node, ast.ImportFrom):
                        raw_imports.append((node.module, file_rel))
            except Exception as exc:
                logger = create_logger(Session.create_session(), __name__)
                logger.error('Failed on file: %s' % file_rel)
                raise exc

    for lib, file in raw_imports:
        name = lib.split('.')[0]
        try:
            if name in _mapping:
                name = _mapping[name]

            version = pkg_resources.get_distribution(name).version
            res.append((name, version))
        except Exception:
            pass

    return res


def _read_requirements(file: str):
    res = []
    for line in read_lines(file):
        if line == '':
            continue
        name, rel, ver = None, None, None
        if '>=' in line:
            rel = '>='
        elif '==' in line:
            rel = '=='

        name = line.split(rel)[0].strip()
        if rel:
            ver = line.split(rel)[1].strip()

        res.append([name, rel, ver])

    return res


def _write_requirements(file: str, reqs: List):
    with open(file, 'w') as f:
        text = '\n'.join(
            [f'{name}{rel}{ver}' if rel else name for name, rel, ver in reqs]
        )

        f.write(text)


def control_requirements(
    path: str, files: List[str] = None, exclude_patterns: List[str] = None
):
    req_file = os.path.join(path, 'requirements.txt')
    if not os.path.exists(req_file):
        with open(req_file, 'w') as f:
            f.write('')

    req_ignore_file = os.path.join(path, 'requirements.ignore.txt')
    if not os.path.exists(req_ignore_file):
        with open(req_ignore_file, 'w') as f:
            f.write('')

    libs = find_imports(path, files=files, exclude_patterns=exclude_patterns)
    module_folder = os.path.dirname(__file__)
    stdlib_file = os.path.join(module_folder, 'req_stdlib')
    ignore_libs = set(read_lines(req_ignore_file) + read_lines(stdlib_file))

    reqs = _read_requirements(req_file)
    for lib, version in libs:
        if lib in ignore_libs:
            continue
        found = False
        for i in range(len(reqs)):
            if reqs[i][0] == lib:
                found = True
                reqs[i][1] = '=='
                reqs[i][2] = version
                break
        if not found:
            reqs.append([lib, '==', version])

    _write_requirements(req_file, reqs)
    return reqs


if __name__ == '__main__':
    folder = '/home/light/projects/mlcomp/'
    # for l, v in find_imports(folder):
    #     print(f'{l}={v}')
    control_requirements(folder, exclude_patterns=['mlcomp/server/front'])
